import {
  useCallback, useEffect, useReducer, useState, useMemo,
} from 'react';
import { useSDK } from '@metamask/sdk-react';
import { Web3 } from 'web3';
import Button from 'react-bootstrap/Button';
import Tab from 'react-bootstrap/Tab';
import Tabs from 'react-bootstrap/Tabs';
import Container from 'react-bootstrap/Container';
import Row from 'react-bootstrap/Row';
import Col from 'react-bootstrap/Col';
import Stack from 'react-bootstrap/Stack';

import { getContractByChainId } from './contract-config';
import Web3Context from './Web3Context';
import DAOArtifact from './assets/DAO.json';
import work1Artifact from './assets/Work1.json';
import AvailabilityStakes from './AvailabilityStakes';
import WorkRequests from './WorkRequests';

const updateList = (list, action) => {
  switch (action.type) {
    case 'update': {
      const newList = [...list];
      newList[Number(action.item.id)] = action.item;
      return newList;
    }
    case 'refresh':
    default:
      return [];
  }
};

const useList = (initialValue) => useReducer(updateList, initialValue ?? []);

function App() {
  const {
    sdk, connected, provider, chainId, account, balance,
  } = useSDK();

  const [DAO, setDAO] = useState();
  const [work1, setWork1] = useState();
  const [work1Price, setWork1Price] = useState();
  const [balanceEther, setBalanceEther] = useState();
  const [reputation, setReputation] = useState();
  const [totalReputation, setTotalReputation] = useState();
  const [posts, dispatchPost] = useList();
  const [validationPools, dispatchValidationPool] = useList();
  const [availabilityStakes, dispatchAvailabilityStake] = useList();
  const [workRequests, dispatchWorkRequest] = useList();

  const web3ProviderValue = useMemo(() => ({
    provider,
    DAO,
    work1,
    work1Price,
    reputation,
    setReputation,
    account,
    chainId,
    availabilityStakes,
    workRequests,
  }), [
    provider, DAO, work1, work1Price, reputation, setReputation, account, chainId,
    availabilityStakes, workRequests,
  ]);

  // In this effect, we initialize everything and add contract event listeners.
  useEffect(() => {
    if (!provider || !chainId || !account || balance === undefined) return;
    const DAOAddress = getContractByChainId(chainId, 'DAO');
    const Work1Address = getContractByChainId(chainId, 'Work1');
    const web3 = new Web3(provider);
    const DAOContract = new web3.eth.Contract(DAOArtifact.abi, DAOAddress);
    const work1Contract = new web3.eth.Contract(work1Artifact.abi, Work1Address);

    /* -------------------------------------------------------------------------------- */
    /* --------------------------- BEGIN FETCHERS ------------------------------------- */
    /* -------------------------------------------------------------------------------- */

    const fetchPrice = async () => {
      const fetchedPrice = await work1Contract.methods.price().call();
      setWork1Price(web3.utils.fromWei(fetchedPrice, 'ether'));
    };

    const fetchReputation = async () => {
      setReputation(await DAOContract.methods.balanceOf(account).call());
      setTotalReputation(await DAOContract.methods.totalSupply().call());
    };

    const fetchPost = async (postIndex) => {
      const p = await DAOContract.methods.posts(postIndex).call();
      p.id = Number(p.id);
      dispatchPost({ type: 'update', item: p });
      return p;
    };

    const fetchPosts = async () => {
      const count = await DAOContract.methods.postCount().call();
      const promises = [];
      dispatchPost({ type: 'refresh' });
      for (let i = 0; i < count; i += 1) {
        promises.push(fetchPost(i));
      }
      await Promise.all(promises);
    };

    const fetchValidationPool = async (poolIndex) => {
      const getPoolStatus = (pool) => {
        if (pool.resolved) {
          return pool.outcome ? 'Accepted' : 'Rejected';
        }
        return pool.timeRemaining > 0 ? 'In Progress' : 'Ready to Evaluate';
      };
      const pool = await DAOContract.methods.validationPools(poolIndex).call();
      pool.id = Number(pool.id);
      pool.timeRemaining = new Date(Number(pool.endTime) * 1000) - new Date();
      pool.status = getPoolStatus(pool);
      dispatchValidationPool({ type: 'update', item: pool });

      // When remaing time expires, we want to update the status for this pool
      if (pool.timeRemaining > 0) {
        setTimeout(() => {
          pool.timeRemaining = 0;
          pool.status = getPoolStatus(pool);
          dispatchValidationPool({ type: 'update', item: pool });
        }, pool.timeRemaining);
      }
    };

    const fetchValidationPools = async () => {
      // TODO: Pagination
      // TODO: Memoization
      // TODO: Caching
      const count = await DAOContract.methods.validationPoolCount().call();
      const promises = [];
      dispatchValidationPool({ type: 'refresh' });
      for (let i = 0; i < count; i += 1) {
        promises.push(fetchValidationPool(i));
      }
      await Promise.all(promises);
    };

    const fetchAvailabilityStake = async (stakeIndex) => {
      const s = await work1Contract.methods.stakes(stakeIndex).call();
      Object.assign(s, {
        id: Number(stakeIndex),
        currentUserIsWorker: () => s.worker.toLowerCase() === account.toString().toLowerCase(),
        timeRemaining: new Date(Number(s.endTime) * 1000) - new Date(),
      });
      dispatchAvailabilityStake({ type: 'update', item: s });
      if (s.timeRemaining > 0) {
        setTimeout(() => {
          s.timeRemaining = 0;
          dispatchAvailabilityStake({ type: 'update', item: s });
        }, s.timeRemaining);
      }
      return s;
    };

    const fetchAvailabilityStakes = async () => {
      const count = await work1Contract.methods.stakeCount().call();
      const promises = [];
      dispatchAvailabilityStake({ type: 'refresh' });
      for (let i = 0; i < count; i += 1) {
        promises.push(fetchAvailabilityStake(i));
      }
      await Promise.all(promises);
    };

    const fetchWorkRequest = async (requestIndex) => {
      const r = await work1Contract.methods.requests(requestIndex).call();
      Object.assign(r, {
        id: Number(requestIndex),
        feeEther: web3.utils.fromWei(r.fee, 'ether'),
        currentUserIsCustomer: () => r.customer.toLowerCase()
            === account.toString().toLowerCase(),
      });
      dispatchWorkRequest({ type: 'update', item: r });
      return r;
    };

    const fetchWorkRequests = async () => {
      const count = await work1Contract.methods.requestCount().call();
      const promises = [];
      dispatchWorkRequest({ type: 'refresh' });
      for (let i = 0; i < count; i += 1) {
        promises.push(fetchWorkRequest(i));
      }
      await Promise.all(promises);
    };

    /* -------------------------------------------------------------------------------- */
    /* --------------------------- END FETCHERS --------------------------------------- */
    /* -------------------------------------------------------------------------------- */

    fetchPrice();
    fetchReputation();
    fetchPosts();
    fetchValidationPools();
    fetchAvailabilityStakes();
    fetchWorkRequests();

    setWork1(work1Contract);
    setDAO(DAOContract);

    /* -------------------------------------------------------------------------------- */
    /* --------------------------- BEGIN EVENT HANDLERS ------------------------------- */
    /* -------------------------------------------------------------------------------- */

    DAOContract.events.PostAdded({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: post added', event);
      fetchPost(event.returnValues.postIndex);
    });

    DAOContract.events.ValidationPoolInitiated({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: validation pool initiated', event);
      fetchValidationPool(event.returnValues.poolIndex);
    });

    DAOContract.events.ValidationPoolResolved({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: validation pool resolved', event);
      fetchReputation();
      fetchValidationPool(event.returnValues.poolIndex);
    });

    work1Contract.events.AvailabilityStaked({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: availability staked', event);
      fetchAvailabilityStake(event.returnValues.stakeIndex);
      fetchReputation();
    });

    work1Contract.events.WorkAssigned({ fromBlock: 'latest' }).on('data', async (event) => {
      console.log('event: work assigned', event);
      const r = await fetchWorkRequest(event.returnValues.requestIndex);
      console.log('work request', r);
      fetchAvailabilityStake(r.stakeIndex);
    });

    work1Contract.events.WorkEvidenceSubmitted({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: work evidence submitted', event);
      fetchWorkRequest(event.returnValues.requestIndex);
    });

    work1Contract.events.WorkApprovalSubmitted({ fromBlock: 'latest' }).on('data', (event) => {
      console.log('event: work approval submitted', event);
      fetchWorkRequest(event.returnValues.requestIndex);
    });
  }, [provider, account, chainId, balance, setReputation, dispatchAvailabilityStake,
    dispatchValidationPool, dispatchWorkRequest, dispatchPost]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- END MAIN INITIALIZION EFFECT ----------------------- */
  /* -------------------------------------------------------------------------------- */

  useEffect(() => {
    if (!provider || balance === undefined) return;
    const web3 = new Web3(provider);
    setBalanceEther(web3.utils.fromWei(balance, 'ether'));
  }, [provider, balance]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- BEGIN UI ACTIONS ----------------------------------- */
  /* -------------------------------------------------------------------------------- */

  const connect = useCallback(async () => {
    try {
      await sdk?.connect();
    } catch (err) {
      console.warn('failed to connect..', err);
    }
  }, [sdk]);

  const disconnect = useCallback(async () => {
    try {
      sdk?.terminate();
    } catch (err) {
      console.warn('failed to disconnect..', err);
    }
  }, [sdk]);

  const watchReputationToken = useCallback(async () => {
    await provider.request({
      method: 'wallet_watchAsset',
      params: {
        type: 'ERC20',
        options: {
          address: getContractByChainId(chainId, 'DAO'),
        },
      },
    });
  }, [provider, chainId]);

  const addPost = useCallback(async () => {
    await DAO.methods.addPost(account).send({
      from: account,
      gas: 1000000,
    });
  }, [DAO, account]);

  const initiateValidationPool = useCallback(async (postIndex, poolDuration) => {
    await DAO.methods.initiateValidationPool(postIndex, poolDuration ?? 3600).send({
      from: account,
      gas: 1000000,
      value: 100,
    });
  }, [DAO, account]);

  const stake = useCallback(async (poolIndex, amount, inFavor) => {
    console.log(`Attempting to stake ${amount} ${inFavor ? 'for' : 'against'} pool ${poolIndex}`);
    await DAO.methods.stake(poolIndex, amount, inFavor).send({
      from: account,
      gas: 999999,
    });

    // Since this is the result we expect from the server, we preemptively set it here.
    // We can let this value be negative -- this would just mean we'll be getting
    // at least one error from the server, and a corrected reputation.
    setReputation((current) => current - BigInt(amount));
  }, [DAO, account, setReputation]);

  const stakeHalfInFavor = useCallback(async (poolIndex) => {
    await stake(poolIndex, reputation / BigInt(2), true);
  }, [stake, reputation]);

  const evaluateOutcome = useCallback(async (poolIndex) => {
    await DAO.methods.evaluateOutcome(poolIndex).send({
      from: account,
      gas: 1000000,
    });
  }, [DAO, account]);

  /* -------------------------------------------------------------------------------- */
  /* --------------------------- END UI ACTIONS ------------------------------------- */
  /* -------------------------------------------------------------------------------- */

  return (
    <Web3Context.Provider value={web3ProviderValue}>
      {!connected && <Button onClick={() => connect()}>Connect</Button>}

      {connected && (
        <>
          <Container>
            <Row>
              <Col>
                {chainId !== '0xaa36a7' && (
                <div>
                  Please switch MetaMask to Sepolia testnet!
                </div>

                )}
              </Col>
            </Row>
            <Row>
              <Col>
                <Stack>
                  <div>
                    {chainId && `Chain ID: ${chainId}`}
                  </div>
                  <div>
                    {`Account: ${account}`}
                  </div>
                  <div>
                    {`Balance: ${balanceEther} ETH`}
                  </div>
                </Stack>
              </Col>
              <Col>
                <Stack>
                  <div>
                    {`Your REP: ${reputation?.toString()}`}
                  </div>
                  <div>
                    {`Total REP: ${totalReputation?.toString()}`}
                  </div>
                  <div>
                    <Button onClick={() => disconnect()}>Disconnect</Button>
                    <Button onClick={() => watchReputationToken()}>Watch REP in MetaMask</Button>
                  </div>
                </Stack>
              </Col>
            </Row>
          </Container>
          <Tabs>
            <Tab eventKey="admin" title="Admin">
              <div>
                {`Posts count: ${posts.length}`}
              </div>
              <div>
                <table className="table">
                  <thead>
                    <tr>
                      <th>ID</th>
                      <th>Author</th>
                      <th>Sender</th>
                      <th>Actions</th>
                    </tr>
                  </thead>
                  <tbody>
                    {posts.filter((x) => !!x).map((post) => (
                      <tr key={post.id}>
                        <td>{post.id.toString()}</td>
                        <td>{post.author}</td>
                        <td>{post.sender}</td>
                        <td>
                          Initiate Validation Pool
                          {' '}
                          <Button onClick={() => initiateValidationPool(post.id, 60)}>
                            1 Min.
                          </Button>
                          {' '}
                          <Button onClick={() => initiateValidationPool(post.id, 3600)}>
                            1 Hr.
                          </Button>
                          {' '}
                          <Button onClick={() => initiateValidationPool(post.id, 86400)}>
                            1 Day
                          </Button>
                        </td>
                      </tr>
                    ))}
                  </tbody>
                </table>
              </div>
              <div>
                <Button onClick={() => addPost()}>Add Post</Button>
              </div>
              <div>
                {`Validation Pool Count: ${validationPools.length}`}
              </div>
              <div>
                <table className="table">
                  <thead>
                    <tr>
                      <th>ID</th>
                      <th>Post ID</th>
                      <th>Fee</th>
                      <th>Duration</th>
                      <th>End Time</th>
                      <th>
                        Stake
                        <br />
                        Count
                      </th>
                      <th>Status</th>
                      <th>Actions</th>
                    </tr>
                  </thead>
                  <tbody>
                    {validationPools.filter((x) => !!x).map((pool) => (
                      <tr key={pool.id}>
                        <td>{pool.id.toString()}</td>
                        <td>{pool.postIndex.toString()}</td>
                        <td>{pool.fee.toString()}</td>
                        <td>{pool.duration.toString()}</td>
                        <td>{new Date(Number(pool.endTime) * 1000).toLocaleString()}</td>
                        <td>{pool.stakeCount.toString()}</td>
                        <td>{pool.status}</td>
                        <td>
                          {!pool.resolved && reputation > 0 && pool.timeRemaining > 0 && (
                          <>
                            <Button onClick={() => stakeHalfInFavor(pool.id)}>
                              Stake 1/2 REP
                            </Button>
                            {' '}
                          </>
                          )}
                          {!pool.resolved && pool.timeRemaining <= 0 && (
                          <Button onClick={() => evaluateOutcome(pool.id)}>
                            Evaluate Outcome
                          </Button>
                          )}
                        </td>
                      </tr>
                    ))}
                  </tbody>
                </table>
              </div>
              <div>
                <h2>Work Contract 1</h2>
                <div>
                  {`Price: ${work1Price} ETH`}
                </div>
                <AvailabilityStakes showActions={false} />
                <WorkRequests />
              </div>
            </Tab>
            <Tab eventKey="worker" title="Worker">
              <h2>Work Contract 1</h2>
              <div>
                {`Price: ${work1Price} ETH`}
              </div>
              <AvailabilityStakes />
              <WorkRequests />
            </Tab>
            <Tab eventKey="customer" title="Customer">
              <h2>Work Contract 1</h2>
              <div>
                {`Price: ${work1Price} ETH`}
              </div>
              <AvailabilityStakes showActions={false} showAmount={false} onlyShowAvailable />
              <WorkRequests showRequestWork />
            </Tab>
          </Tabs>
        </>
      )}
    </Web3Context.Provider>
  );
}

export default App;
